#include <hpgcc49.h>
#include <hpgraphics.h>

#define WINDOW_WIDTH  				131
#define WINDOW_HEIGHT  				80
#define NULL 						0
#define PI         					((float) 3.141592654)
#define DEG_TO_RAD(ang) 			((ang) * PI / 180.0)
#define EPSILON_E5 					(float) (1E-5)
#define EPSILON_E3 					(float) (1E-3)

#define pi                          PI

#define COL_H						6

#define COL_MIN						COL_H * 0
#define COL_MAX						COL_H * 1
#define COL_STEP					COL_H * 2
#define COL_ISO						COL_H * 3
#define COL_PRECISION				COL_H * 4
#define COL_SCALE					COL_H * 5

#define COL_A						COL_H * 6
#define COL_B 						COL_H * 7
#define COL_C						COL_H * 8
#define COL_D						COL_H * 9
#define COL_E						COL_H * 10
#define COL_F						COL_H * 11

#define INC 						0.5
#define INC_VAR 					0.01
#define INC_PRE 					0.2

//																								3d engine
typedef struct {
	float x,y,z;
} VECTOR3D, POINT3D, *VECTOR3D_PTR, *POINT3D_PTR;

typedef struct {
	POINT3D p0; 									
	VECTOR3D n; 									
} PLANE, *PLANE_PTR;

typedef struct {
	union {
		float M[4][4];
		struct {
			float M00, M01, M02, M03;
			float M10, M11, M12, M13;
			float M20, M21, M22, M23;
			float M30, M31, M32, M33;
		};
	};
} MATRICE, *MATRICE_PTR;

typedef struct {
	union {
		float M[4]; 
		struct { float x,y,z,w;};
	};
} VECTOR4D, POINT4D, *VECTOR4D_PTR, *POINT4D_PTR;

typedef struct {
	union {     
		struct {float x,y,z,w;};    
		struct  {POINT4D  v;};             
	}; 
} VERTEX, *VERTEX_PTR;

typedef struct {					 
	int cull;
	int vert[3];
} POLYGONE, *POLYGONE_PTR;

typedef struct {   
	POINT4D world_pos;    
	VERTEX_PTR vlist;
	POLYGONE_PTR plist;  
	POLYGONE_PTR *pplist;  
} OBJECT, *OBJECT_PTR;

typedef struct {   					
	POINT4D pos;    								
	VECTOR4D dir;  	 							
	float view_dist;  								
	float fov;        							
	float near_z;    							
	float far_z;      					
	PLANE rt_plane;  					
	PLANE lt_plane;  						
	PLANE tp_plane;  						
	PLANE bt_plane;  						              
	float viewplane_width;     					
	float viewplane_height;  					  	
	float viewport_width; 						   
	float viewport_height;
	float viewport_center_x;  						
	float viewport_center_y;
	float aspect_ratio;
	MATRICE mcam;   								
	MATRICE mper; 							    
	MATRICE mscr; 							    
} CAMERA, *CAMERA_PTR;


//																								constants, macros, prototypes
const MATRICE ID_MAT = {1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1};

#define MAT_IDENTITY(m) {memcpy((void *)(m), (void *)&ID_MAT, sizeof(MATRICE));}
#define MAT_COPY(src_mat, dest_mat) {memcpy((void *)(dest_mat), (void *)(src_mat), sizeof(MATRICE) ); }
#define SWAP(a,b,t) {t=a; a=b; b=t;}

int compare_polygone(const void *arg1, const void *arg2);
void init_plane(PLANE_PTR plane, POINT3D_PTR p0, VECTOR3D_PTR normal);
void init_matrice(MATRICE_PTR ma, 
				float m00, float m01, float m02, float m03, float m10, float m11, float m12, float m13,
				float m20, float m21, float m22, float m23, float m30, float m31, float m32, float m33);				
void multiply_matrice(MATRICE_PTR ma, MATRICE_PTR mb, MATRICE_PTR mprod);
void build_rotation_matrice(float theta_x, float theta_y, float theta_z, MATRICE_PTR mrot);
void init_camera(CAMERA_PTR cam, POINT4D_PTR cam_pos, VECTOR4D_PTR cam_dir, float near_z, float far_z, 
					float fov, float viewport_width,  float viewport_height);

//																								globals			
CAMERA cam; POINT4D cam_pos = { 0, 0, 0, 1 }; VECTOR4D cam_dir = { 0, 0, 0, 1 };
OBJECT surface; 
int raster_x[3], raster_y[3];
MATRICE mrot; float x_angle = 10.0; float y_angle = 0.0; float z_angle = 0.0;
int polygone, vertex; POINT4D presult; int col, row; float sum;
int num_verts, num_polys;
int display = 0;
VECTOR3D *vertex_array;
char buffer[100];

//																								marching cube
typedef struct {
	union {
		struct {
			float min, max, step, isovalue, precision, scale;
		};
		float p[5];	
	};
} PARAMETER;

typedef struct {
	VECTOR3D p[8];
	float val[8];
} CUBE;


int polygonise();
int compute_marching_cube();
CUBE cube;
VECTOR3D cube_edge[12];
PARAMETER parameters = {-3.0, 3.0, 9.0, 1.0, 0.2, 2.0};
int interpolate_vertex[12][2] = {{0,1}, {1,2}, {2,3}, {3,0}, {4,5}, {5,6}, {6,7}, {7,4}, {0,4}, {1,5}, {2,6}, {3,7}};
int cube_index, index_v, idx, vertex_exists, edge_index, cube_vertex, token, interpolate;
int flag_compute_marching_cube = 1;
float interpolate_delta = 0.0001;

//																								equation
#define isnumber(c)  			((c >= '0' && c <= '9') || c == '.')
#define isvariable(c)  			(c == 'a' || c == 'b' || c == 'c' || c == 'd' || c == 'e' || c == 'f')
#define isunknown(c)			(c == 'x' || c == 'y' || c == 'z')
#define isoperator(c)			(c == 131)

#define MAX_TOKEN				300
#define MAX_VARIABLE			6
#define MAX_PARAMETER			6

#define MAX_FUNCTION			21

#define STACK_DEPTH				100
#define NAME_LENGTH				20

#define NUMBER_TYPE				0
#define FUNCTION_TYPE			1
#define VARIABLE_TYPE			2
#define UNKNOWN_TYPE			3
#define OPERATOR_TYPE			4

#define	X						0
#define	Y						1
#define	Z						2

#define	A						0
#define	B						1
#define	C						2
#define	D						3
#define E						4
#define	F						5

typedef struct {
	float value[MAX_VARIABLE];
	int flag[MAX_VARIABLE];
} VARIABLE;

typedef struct {
	int type;
	union {
		float value;
		int index;
	};
} TOKEN; 

typedef struct {
	char *name;
	int argnum;
	double (*function)();
} FUNCTION; 

typedef struct {
	char *expression;
	char *parameter;
	char *variable;
	TOKEN xyz_token[MAX_TOKEN];
	int xyz_tokennumber;
} EQUATION; 

typedef struct {
	char *xyz;
} EQUATION_INFO;

int parse_expression();
void get_parameters(), get_variable();
int get_equation();

double addition(), subtraction();
double multiplication(), division();
double sinhp(), coshp(), tanhp();
double neg(), inv();
double sq();

double atof_local(const char * _buf);

EQUATION equation;
EQUATION_INFO equation_info;
VARIABLE variable = {1.0, 1.0, 1.0, 1.0, 1.0, 1.0};
int stack_index;
char function_name[NAME_LENGTH];
float execution_stack[STACK_DEPTH];

int error_memory(char *message);

//																								hash table
#define MAX_HASH_VERTEX	 	5
#define HASH_TABLE_SIZE		1009
#define MAX_HASH_INDEX		4500

typedef struct {
    float x,y,z;
    int num_hash_vertex;
    int a,b,c;
} HASH_VERTEX;

typedef struct {
    HASH_VERTEX vertex[MAX_HASH_VERTEX];
    int num_cell;
} HASH_CELL;

HASH_CELL *hash_table;
HASH_VERTEX **polys_vertex_index;
int num_poly_verts;

FUNCTION function[] = {
	{"SIN", 1, sin},
	{"COS", 1, cos},
	{"TAN", 1, tan},
	{"SINH", 1, sinhp},
	{"COSH", 1, coshp},
	{"TANH", 1, tanhp},
	{"ASIN", 1, asin},
	{"ACOS", 1, acos},
	{"ATAN", 1, atan},
	{"LN", 1, log},
	{"LOG", 1, log10},
	{"NEG", 1, neg},
	{"INV", 1, inv},
	{"ABS", 1, fabs},
	{"EXP", 1, exp},
	{"SQ", 1, sq},
	{"^", 2, pow},
	{"+", 2, addition},
	{"-", 2, subtraction},
	{"*", 2, multiplication},
	{"/", 2, division},
};

//																								functions
int main() {	
	hpg_clear(); if(!get_equation()) return 0;
	hpg_set_mode_gray16(1); hpg_clear(); hpg_flip(); hpg_clear();		
	sys_slowOff(); if(!compute_marching_cube()) return 0;
	init_camera(&cam, &cam_pos, &cam_dir, 10.0, 50.0, 90.0, WINDOW_WIDTH, WINDOW_HEIGHT);
	surface.world_pos.x = 0; surface.world_pos.y = 0; surface.world_pos.z = 20.0; surface.world_pos.w = 1.0;
    while (!keyb_isON()) {	
		y_angle += 0.02;
		if (keyb_isAnyKeyPressed()) {		
			if(keyb_isUp()) x_angle += 0.1; else if(keyb_isDown()) x_angle += -0.1;
			else if(keyb_isLeft()) y_angle += 0.1; else if(keyb_isRight()) y_angle += -0.1;
			else if(keyb_isKeyPressed(4,0)) {z_angle += 0.1;}
			else if(keyb_isKeyPressed(3,0)) {z_angle -= 0.1;}
				
			else if(keyb_isKeyPressed(0,5)) {if(surface.world_pos.z > 10.0)  surface.world_pos.z -= 1.0;}
			else if(keyb_isKeyPressed(0,4)) {if(surface.world_pos.z < 50.0)  surface.world_pos.z += 1.0;}
			else if(keyb_isKeyPressed(1,4)) {if(surface.world_pos.x < 20.0)  surface.world_pos.x += 1.0;} 
			else if(keyb_isKeyPressed(3,4)) {if(surface.world_pos.x > -20.0) surface.world_pos.x -= 1.0;} 
			else if(keyb_isKeyPressed(2,3)) {if(surface.world_pos.y < 20.0)  surface.world_pos.y += 1.0;}
			else if(keyb_isKeyPressed(2,5)) {if(surface.world_pos.y > -25.0) surface.world_pos.y -= 1.0;}
			
			else if (keyb_isLS()) { 
				if(keyb_isKeyPressed(6,4)) {if(parameters.scale > INC) parameters.scale -= INC;}	
				else if(keyb_isKeyPressed(1,1)){parameters.min -= INC;compute_marching_cube();}
				else if(keyb_isKeyPressed(0,1)){parameters.max -= INC;compute_marching_cube();}
				else if(keyb_isKeyPressed(7,1)){parameters.step -= INC;compute_marching_cube();}	
				else if(keyb_isKeyPressed(6,6)){parameters.isovalue -= INC;compute_marching_cube();}
				else if(keyb_isKeyPressed(7,0)){parameters.precision -= INC_PRE;if(parameters.precision <= 0) parameters.precision = 0; compute_marching_cube();}
				
				else if(keyb_isKeyPressed(5,0)){if(variable.flag[A]){variable.value[A] -= INC_VAR; compute_marching_cube();}}
				else if(keyb_isKeyPressed(5,1)){if(variable.flag[B]){variable.value[B] -= INC_VAR; compute_marching_cube();}}
				else if(keyb_isKeyPressed(5,2)){if(variable.flag[C]){variable.value[C] -= INC_VAR; compute_marching_cube();}}
				else if(keyb_isKeyPressed(5,3)){if(variable.flag[D]){variable.value[D] -= INC_VAR; compute_marching_cube();}}
				else if(keyb_isKeyPressed(5,4)){if(variable.flag[E]){variable.value[E] -= INC_VAR; compute_marching_cube();}}
				else if(keyb_isKeyPressed(5,5)){if(variable.flag[F]){variable.value[F] -= INC_VAR; compute_marching_cube();}}
		}
			
			else if(keyb_isKeyPressed(6,4)){parameters.scale += INC; }
			else if(keyb_isKeyPressed(1,1)){parameters.min += INC;compute_marching_cube();}
			else if(keyb_isKeyPressed(0,1)){parameters.max += INC;compute_marching_cube();}
			else if(keyb_isKeyPressed(7,1)){parameters.step += INC;compute_marching_cube();}	
			else if(keyb_isKeyPressed(6,6)){parameters.isovalue += INC;compute_marching_cube();}
			else if(keyb_isKeyPressed(7,0)){parameters.precision += INC_PRE;compute_marching_cube();}
			
			else if(keyb_isKeyPressed(5,0)){if(variable.flag[A]){variable.value[A] += INC_VAR; compute_marching_cube();}}
			else if(keyb_isKeyPressed(5,1)){if(variable.flag[B]){variable.value[B] += INC_VAR; compute_marching_cube();}}
			else if(keyb_isKeyPressed(5,2)){if(variable.flag[C]){variable.value[C] += INC_VAR; compute_marching_cube();}}
			else if(keyb_isKeyPressed(5,3)){if(variable.flag[D]){variable.value[D] += INC_VAR; compute_marching_cube();}}
			else if(keyb_isKeyPressed(5,4)){if(variable.flag[E]){variable.value[E] += INC_VAR; compute_marching_cube();}}
			else if(keyb_isKeyPressed(5,5)){if(variable.flag[F]){variable.value[F] += INC_VAR; compute_marching_cube();}}

//																								freeze the screen
			else if(keyb_isKeyPressed(4,1)) keyb_getkey(1);
			else if(keyb_isKeyPressed(5,6)) { display = display == 1 ? 0 : 1; }			

			else if (keyb_isKeyPressed(6,5)) {
				hpg_clear(); 
				sprintf(buffer, "num. vertices:..%d", num_verts); hpg_draw_text(buffer, 0, COL_H * 0);
				sprintf(buffer, "num. polygons:..%d", num_polys); hpg_draw_text(buffer, 0, COL_H * 1);	
				hpg_draw_text("F= ",0 ,COL_H * 3);hpg_draw_text(equation_info.xyz + 1, 12, COL_H * 3);
				hpg_flip(); keyb_getkey(1); 	
			}	
		}
		
		if(!flag_compute_marching_cube) return 0;
		
		for( vertex = 0; vertex < num_verts; vertex++) {			
			surface.vlist[vertex].x = vertex_array[vertex].x * parameters.scale; surface.vlist[vertex].y = vertex_array[vertex].y * parameters.scale;
			surface.vlist[vertex].z = vertex_array[vertex].z * parameters.scale; surface.vlist[vertex].w = 1.0;
		}
		
		for (polygone = 0; polygone < num_polys; polygone++) {surface.pplist[polygone]->cull = 0;}
		
		build_rotation_matrice(x_angle, y_angle, z_angle, &mrot);	

		for (vertex = 0; vertex < num_verts; vertex++) {
			for (col = 0; col < 4; col++) {       						
				sum = 0; 
				for (row = 0; row < 4; row++) { sum += (surface.vlist[vertex].v.M[row] * mrot.M[row][col]);} 
				presult.M[col] = sum;
			} 
			surface.vlist[vertex].v.x = presult.x + surface.world_pos.x; surface.vlist[vertex].v.y = presult.y + surface.world_pos.y;
			surface.vlist[vertex].v.z = presult.z + surface.world_pos.z; surface.vlist[vertex].v.w = 1;	
			for (col = 0; col < 4; col++) {       						
				sum = 0; 
				for (row = 0; row < 4; row++) { sum += (surface.vlist[vertex].v.M[row] * cam.mcam.M[row][col]);} 
				presult.M[col] = sum;
			} 
			surface.vlist[vertex].v.x = presult.x; surface.vlist[vertex].v.y = presult.y;
			surface.vlist[vertex].v.z = presult.z; surface.vlist[vertex].v.w = presult.w;	
		}
		
		qsort((void *) surface.pplist, num_polys, sizeof(POLYGONE_PTR), compare_polygone);
	
		for (polygone = 0; polygone < num_polys; polygone++) {
			if(surface.vlist[surface.pplist[polygone]->vert[0]].z <= EPSILON_E3 || surface.vlist[surface.pplist[polygone]->vert[1]].z <= EPSILON_E3
				|| surface.vlist[surface.pplist[polygone]->vert[2]].z <= EPSILON_E3) 
			surface.pplist[polygone]->cull = 1;
		}

		float alpha = (0.5 * cam.viewport_width - 0.5); float beta = (0.5 * cam.viewport_height - 0.5);
		for (vertex = 0; vertex < num_verts; vertex++) {
			float z = surface.vlist[vertex].z; if(z == 0.0) z = EPSILON_E5;	
			surface.vlist[vertex].x = cam.view_dist * surface.vlist[vertex].x / z;
			surface.vlist[vertex].y = cam.view_dist * surface.vlist[vertex].y * cam.aspect_ratio / z;
			surface.vlist[vertex].x = alpha + alpha * surface.vlist[vertex].x;
			surface.vlist[vertex].y = beta - beta * surface.vlist[vertex].y;
		}

		hpg_clear();
		for (polygone = 0; polygone < num_polys; polygone++) {
			if(surface.pplist[polygone]->cull == 1) continue;
			raster_x[0] = (int) surface.vlist[surface.pplist[polygone]->vert[0]].x;
			raster_y[0] = (int) surface.vlist[surface.pplist[polygone]->vert[0]].y;
			raster_x[1] = (int) surface.vlist[surface.pplist[polygone]->vert[1]].x;
			raster_y[1] = (int) surface.vlist[surface.pplist[polygone]->vert[1]].y;
			raster_x[2] = (int) surface.vlist[surface.pplist[polygone]->vert[2]].x;
			raster_y[2] = (int) surface.vlist[surface.pplist[polygone]->vert[2]].y;
			hpg_set_color(hpg_stdscreen,HPG_COLOR_GRAY_4); hpg_fill_polygon(raster_x, raster_y, 3);	
			hpg_set_color(hpg_stdscreen, HPG_COLOR_BLACK); hpg_draw_polygon(raster_x, raster_y, 3);			
		}
		
		if(display) {
			sprintf(buffer, "min(t)...%.1f", parameters.min); hpg_draw_text(buffer, 0, COL_MIN);
			sprintf(buffer, "max(u)...%.1f", parameters.max); hpg_draw_text(buffer, 0, COL_MAX);
			sprintf(buffer, "step(l)..%.1f", parameters.step); hpg_draw_text(buffer, 0, COL_STEP);
			sprintf(buffer, "iso(j)...%.1f", parameters.isovalue); hpg_draw_text(buffer, 0, COL_ISO);
			sprintf(buffer, "preci(k):%.1f", parameters.precision); hpg_draw_text(buffer, 0, COL_PRECISION);
			sprintf(buffer, "scale(h).%.1f", parameters.scale); hpg_draw_text(buffer, 0, COL_SCALE);
			if(variable.flag[A]) {sprintf(buffer, "a..%.2f",  variable.value[A]); hpg_draw_text(buffer, 0, COL_A);}
			if(variable.flag[B]) {sprintf(buffer, "b..%.2f",  variable.value[B]); hpg_draw_text(buffer, 0, COL_B);}
			if(variable.flag[C]) {sprintf(buffer, "c..%.2f",  variable.value[C]); hpg_draw_text(buffer, 0, COL_C);}
			if(variable.flag[D]) {sprintf(buffer, "d..%.2f",  variable.value[D]); hpg_draw_text(buffer, 0, COL_D);}	
			if(variable.flag[E]) {sprintf(buffer, "e..%.2f",  variable.value[E]); hpg_draw_text(buffer, 0, COL_E);}
			if(variable.flag[F]) {sprintf(buffer, "f..%.2f",  variable.value[F]); hpg_draw_text(buffer, 0, COL_F);}	

		}		
		hpg_flip();
	}
	sys_slowOn(); return 1;
}

int compute_marching_cube() {
	num_verts = num_poly_verts = 0;
	free((void *) surface.vlist); free((void *) surface.plist); free((void *) surface.pplist); free((void *) vertex_array);
	if((hash_table = (HASH_CELL *) calloc(HASH_TABLE_SIZE, sizeof(HASH_CELL))) == NULL) return(error_memory("hash_table\n")); 
	if((polys_vertex_index = (HASH_VERTEX **)  malloc(MAX_HASH_INDEX * sizeof(int))) == NULL ) return(error_memory("polys_vertex_index\n"));
	if(!polygonise()) {flag_compute_marching_cube = 0; return 0;} 
	if((surface.plist = (POLYGONE_PTR) malloc((num_poly_verts / 3) * sizeof(POLYGONE))) == NULL ) return(error_memory("plist\n"));
	if((vertex_array = (VECTOR3D *) malloc(num_verts * sizeof(VECTOR3D))) == NULL) return(error_memory("vertex_array\n"));
//																										build the mesh from the hash table
	int i, j, n_v;
	for(i = 0, num_polys = 0, n_v = 0; i < num_poly_verts; i += 3, num_polys++) {
		for(j = 0; j < 3; j++){
			HASH_VERTEX *p_v = polys_vertex_index[i+j];
			if(p_v->num_hash_vertex > 0) {
				vertex_array[n_v].x = p_v->x / p_v->num_hash_vertex; vertex_array[n_v].y = p_v->y / p_v->num_hash_vertex; 
				vertex_array[n_v].z = p_v->z / p_v->num_hash_vertex; p_v->num_hash_vertex = -n_v; n_v++; 
			}
			surface.plist[num_polys].vert[j] = -p_v->num_hash_vertex;
		}
//																										eliminate degenerate triangles
		if( surface.plist[num_polys].vert[0] == surface.plist[num_polys].vert[1] ||
			surface.plist[num_polys].vert[0] == surface.plist[num_polys].vert[2] ||
			surface.plist[num_polys].vert[1] == surface.plist[num_polys].vert[2]) --num_polys;
	}	
	free((void *) hash_table); free((void *) polys_vertex_index);
	if((surface.plist = (POLYGONE_PTR) realloc(surface.plist, num_polys * sizeof(POLYGONE))) == NULL ) return(error_memory("plist\n"));
	if((surface.pplist = (POLYGONE_PTR *) malloc((num_polys) * sizeof(POLYGONE_PTR))) == NULL ) return(error_memory("pplist\n"));
	for(i = 0; i < num_polys; i++) surface.pplist[i] = &surface.plist[i];
	if((surface.vlist = (VERTEX_PTR) malloc(num_verts * sizeof(VERTEX))) == NULL ) return(error_memory("vlist\n"));
	return 1;
}

int error_memory(char *message) {
	hpg_clear(); hpg_draw_text("Memory allocation error!\n", 0, 0); hpg_draw_text(message, 0, 8);
	sprintf(buffer, "vertices:..%d polys:..%d\n", num_verts, num_poly_verts / 3); hpg_draw_text(buffer, 0, 16);
	hpg_flip(); keyb_getkey(1); flag_compute_marching_cube = 0; return 0;
}

int edge_table[256] = {
	0x0  , 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c,
	0x80c, 0x905, 0xa0f, 0xb06, 0xc0a, 0xd03, 0xe09, 0xf00,
	0x190, 0x99 , 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c,
	0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90,
	0x230, 0x339, 0x33 , 0x13a, 0x636, 0x73f, 0x435, 0x53c,
	0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30,
	0x3a0, 0x2a9, 0x1a3, 0xaa , 0x7a6, 0x6af, 0x5a5, 0x4ac,
	0xbac, 0xaa5, 0x9af, 0x8a6, 0xfaa, 0xea3, 0xda9, 0xca0,
	0x460, 0x569, 0x663, 0x76a, 0x66 , 0x16f, 0x265, 0x36c,
	0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60,
	0x5f0, 0x4f9, 0x7f3, 0x6fa, 0x1f6, 0xff , 0x3f5, 0x2fc,
	0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0,
	0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x55 , 0x15c,
	0xe5c, 0xf55, 0xc5f, 0xd56, 0xa5a, 0xb53, 0x859, 0x950,
	0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0xcc ,
	0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0,
	0x8c0, 0x9c9, 0xac3, 0xbca, 0xcc6, 0xdcf, 0xec5, 0xfcc,
	0xcc , 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0,
	0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c,
	0x15c, 0x55 , 0x35f, 0x256, 0x55a, 0x453, 0x759, 0x650,
	0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc,
	0x2fc, 0x3f5, 0xff , 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0,
	0xb60, 0xa69, 0x963, 0x86a, 0xf66, 0xe6f, 0xd65, 0xc6c,
	0x36c, 0x265, 0x16f, 0x66 , 0x76a, 0x663, 0x569, 0x460,
	0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac,
	0x4ac, 0x5a5, 0x6af, 0x7a6, 0xaa , 0x1a3, 0x2a9, 0x3a0,
	0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c,
	0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x33 , 0x339, 0x230,
	0xe90, 0xf99, 0xc93, 0xd9a, 0xa96, 0xb9f, 0x895, 0x99c,
	0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x99 , 0x190,
	0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c,
	0x70c, 0x605, 0x50f, 0x406, 0x30a, 0x203, 0x109, 0x0};

int triangle_table[256][13] = {
	{-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    { 8, 3, 0,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    { 9, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    { 8, 3, 1, 8, 1, 9,-1,-1,-1,-1,-1,-1,-1},
    {10, 1, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    { 8, 3, 0, 1, 2,10,-1,-1,-1,-1,-1,-1,-1},
    { 9, 0, 2, 9, 2,10,-1,-1,-1,-1,-1,-1,-1},
    { 3, 2, 8, 2,10, 8, 8,10, 9,-1,-1,-1,-1},
    {11, 2, 3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    {11, 2, 0,11, 0, 8,-1,-1,-1,-1,-1,-1,-1},
    {11, 2, 3, 0, 1, 9,-1,-1,-1,-1,-1,-1,-1},
    { 2, 1,11, 1, 9,11,11, 9, 8,-1,-1,-1,-1},
    {10, 1, 3,10, 3,11,-1,-1,-1,-1,-1,-1,-1},
    { 1, 0,10, 0, 8,10,10, 8,11,-1,-1,-1,-1},
    { 0, 3, 9, 3,11, 9, 9,11,10,-1,-1,-1,-1},
    { 8,10, 9, 8,11,10,-1,-1,-1,-1,-1,-1,-1},
    { 8, 4, 7,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    { 3, 0, 4, 3, 4, 7,-1,-1,-1,-1,-1,-1,-1},
    { 1, 9, 0, 8, 4, 7,-1,-1,-1,-1,-1,-1,-1},
    { 9, 4, 1, 4, 7, 1, 1, 7, 3,-1,-1,-1,-1},
    {10, 1, 2, 8, 4, 7,-1,-1,-1,-1,-1,-1,-1},
    { 2,10, 1, 0, 4, 7, 0, 7, 3,-1,-1,-1,-1},
    { 4, 7, 8, 0, 2,10, 0,10, 9,-1,-1,-1,-1},
    { 2, 7, 3, 2, 9, 7, 7, 9, 4, 2,10, 9,-1},
    { 2, 3,11, 7, 8, 4,-1,-1,-1,-1,-1,-1,-1},
    { 7,11, 4,11, 2, 4, 4, 2, 0,-1,-1,-1,-1},
    { 3,11, 2, 4, 7, 8, 9, 0, 1,-1,-1,-1,-1},
    { 2, 7,11, 2, 1, 7, 1, 4, 7, 1, 9, 4,-1},
    { 8, 4, 7,11,10, 1,11, 1, 3,-1,-1,-1,-1},
    {11, 4, 7, 1, 4,11, 1,11,10, 1, 0, 4,-1},
    { 3, 8, 0, 7,11, 4,11, 9, 4,11,10, 9,-1},
    { 7,11, 4, 4,11, 9,11,10, 9,-1,-1,-1,-1},
    { 9, 5, 4,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    { 3, 0, 8, 4, 9, 5,-1,-1,-1,-1,-1,-1,-1},
    { 5, 4, 0, 5, 0, 1,-1,-1,-1,-1,-1,-1,-1},
    { 4, 8, 5, 8, 3, 5, 5, 3, 1,-1,-1,-1,-1},
    { 2,10, 1, 9, 5, 4,-1,-1,-1,-1,-1,-1,-1},
    { 0, 8, 3, 5, 4, 9,10, 1, 2,-1,-1,-1,-1},
    {10, 5, 2, 5, 4, 2, 2, 4, 0,-1,-1,-1,-1},
    { 3, 4, 8, 3, 2, 4, 2, 5, 4, 2,10, 5,-1},
    {11, 2, 3, 9, 5, 4,-1,-1,-1,-1,-1,-1,-1},
    { 9, 5, 4, 8,11, 2, 8, 2, 0,-1,-1,-1,-1},
    { 3,11, 2, 1, 5, 4, 1, 4, 0,-1,-1,-1,-1},
    { 8, 5, 4, 2, 5, 8, 2, 8,11, 2, 1, 5,-1},
    { 5, 4, 9, 1, 3,11, 1,11,10,-1,-1,-1,-1},
    { 0, 9, 1, 4, 8, 5, 8,10, 5, 8,11,10,-1},
    { 3, 4, 0, 3,10, 4, 4,10, 5, 3,11,10,-1},
    { 4, 8, 5, 5, 8,10, 8,11,10,-1,-1,-1,-1},
    { 9, 5, 7, 9, 7, 8,-1,-1,-1,-1,-1,-1,-1},
    { 0, 9, 3, 9, 5, 3, 3, 5, 7,-1,-1,-1,-1},
    { 8, 0, 7, 0, 1, 7, 7, 1, 5,-1,-1,-1,-1},
    { 1, 7, 3, 1, 5, 7,-1,-1,-1,-1,-1,-1,-1},
    { 1, 2,10, 5, 7, 8, 5, 8, 9,-1,-1,-1,-1},
    { 9, 1, 0,10, 5, 2, 5, 3, 2, 5, 7, 3,-1},
    { 5, 2,10, 8, 2, 5, 8, 5, 7, 8, 0, 2,-1},
    {10, 5, 2, 2, 5, 3, 5, 7, 3,-1,-1,-1,-1},
    {11, 2, 3, 8, 9, 5, 8, 5, 7,-1,-1,-1,-1},
    { 9, 2, 0, 9, 7, 2, 2, 7,11, 9, 5, 7,-1},
    { 0, 3, 8, 2, 1,11, 1, 7,11, 1, 5, 7,-1},
    { 2, 1,11,11, 1, 7, 1, 5, 7,-1,-1,-1,-1},
    { 3, 9, 1, 3, 8, 9, 7,11,10, 7,10, 5,-1},
    { 9, 1, 0,10, 7,11,10, 5, 7,-1,-1,-1,-1},
    { 3, 8, 0, 7,10, 5, 7,11,10,-1,-1,-1,-1},
    {11, 5, 7,11,10, 5,-1,-1,-1,-1,-1,-1,-1},
    {10, 6, 5,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    { 8, 3, 0,10, 6, 5,-1,-1,-1,-1,-1,-1,-1},
    { 0, 1, 9, 5,10, 6,-1,-1,-1,-1,-1,-1,-1},
    {10, 6, 5, 9, 8, 3, 9, 3, 1,-1,-1,-1,-1},
    { 1, 2, 6, 1, 6, 5,-1,-1,-1,-1,-1,-1,-1},
    { 0, 8, 3, 2, 6, 5, 2, 5, 1,-1,-1,-1,-1},
    { 5, 9, 6, 9, 0, 6, 6, 0, 2,-1,-1,-1,-1},
    { 9, 6, 5, 3, 6, 9, 3, 9, 8, 3, 2, 6,-1},
    { 3,11, 2,10, 6, 5,-1,-1,-1,-1,-1,-1,-1},
    { 6, 5,10, 2, 0, 8, 2, 8,11,-1,-1,-1,-1},
    { 1, 9, 0, 6, 5,10,11, 2, 3,-1,-1,-1,-1},
    { 1,10, 2, 5, 9, 6, 9,11, 6, 9, 8,11,-1},
    {11, 6, 3, 6, 5, 3, 3, 5, 1,-1,-1,-1,-1},
    { 0, 5, 1, 0,11, 5, 5,11, 6, 0, 8,11,-1},
    { 0, 5, 9, 0, 3, 5, 3, 6, 5, 3,11, 6,-1},
    { 5, 9, 6, 6, 9,11, 9, 8,11,-1,-1,-1,-1},
    {10, 6, 5, 4, 7, 8,-1,-1,-1,-1,-1,-1,-1},
    { 5,10, 6, 7, 3, 0, 7, 0, 4,-1,-1,-1,-1},
    { 5,10, 6, 0, 1, 9, 8, 4, 7,-1,-1,-1,-1},
    { 4, 5, 9, 6, 7,10, 7, 1,10, 7, 3, 1,-1},
    { 7, 8, 4, 5, 1, 2, 5, 2, 6,-1,-1,-1,-1},
    { 4, 1, 0, 4, 5, 1, 6, 7, 3, 6, 3, 2,-1},
    { 9, 4, 5, 8, 0, 7, 0, 6, 7, 0, 2, 6,-1},
    { 4, 5, 9, 6, 3, 2, 6, 7, 3,-1,-1,-1,-1},
    { 7, 8, 4, 2, 3,11,10, 6, 5,-1,-1,-1,-1},
    {11, 6, 7,10, 2, 5, 2, 4, 5, 2, 0, 4,-1},
    {11, 6, 7, 8, 0, 3, 1,10, 2, 9, 4, 5,-1},
    { 6, 7,11, 1,10, 2, 9, 4, 5,-1,-1,-1,-1},
    { 6, 7,11, 4, 5, 8, 5, 3, 8, 5, 1, 3,-1},
    { 6, 7,11, 4, 1, 0, 4, 5, 1,-1,-1,-1,-1},
    { 4, 5, 9, 3, 8, 0,11, 6, 7,-1,-1,-1,-1},
    { 9, 4, 5, 7,11, 6,-1,-1,-1,-1,-1,-1,-1},
    {10, 6, 4,10, 4, 9,-1,-1,-1,-1,-1,-1,-1},
    { 8, 3, 0, 9,10, 6, 9, 6, 4,-1,-1,-1,-1},
    { 1,10, 0,10, 6, 0, 0, 6, 4,-1,-1,-1,-1},
    { 8, 6, 4, 8, 1, 6, 6, 1,10, 8, 3, 1,-1},
    { 9, 1, 4, 1, 2, 4, 4, 2, 6,-1,-1,-1,-1},
    { 1, 0, 9, 3, 2, 8, 2, 4, 8, 2, 6, 4,-1},
    { 2, 4, 0, 2, 6, 4,-1,-1,-1,-1,-1,-1,-1},
    { 3, 2, 8, 8, 2, 4, 2, 6, 4,-1,-1,-1,-1},
    { 2, 3,11, 6, 4, 9, 6, 9,10,-1,-1,-1,-1},
    { 0,10, 2, 0, 9,10, 4, 8,11, 4,11, 6,-1},
    {10, 2, 1,11, 6, 3, 6, 0, 3, 6, 4, 0,-1},
    {10, 2, 1,11, 4, 8,11, 6, 4,-1,-1,-1,-1},
    { 1, 4, 9,11, 4, 1,11, 1, 3,11, 6, 4,-1},
    { 0, 9, 1, 4,11, 6, 4, 8,11,-1,-1,-1,-1},
    {11, 6, 3, 3, 6, 0, 6, 4, 0,-1,-1,-1,-1},
    { 8, 6, 4, 8,11, 6,-1,-1,-1,-1,-1,-1,-1},
    { 6, 7,10, 7, 8,10,10, 8, 9,-1,-1,-1,-1},
    { 9, 3, 0, 6, 3, 9, 6, 9,10, 6, 7, 3,-1},
    { 6, 1,10, 6, 7, 1, 7, 0, 1, 7, 8, 0,-1},
    { 6, 7,10,10, 7, 1, 7, 3, 1,-1,-1,-1,-1},
    { 7, 2, 6, 7, 9, 2, 2, 9, 1, 7, 8, 9,-1},
    { 1, 0, 9, 3, 6, 7, 3, 2, 6,-1,-1,-1,-1},
    { 8, 0, 7, 7, 0, 6, 0, 2, 6,-1,-1,-1,-1},
    { 2, 7, 3, 2, 6, 7,-1,-1,-1,-1,-1,-1,-1},
    { 7,11, 6, 3, 8, 2, 8,10, 2, 8, 9,10,-1},
    {11, 6, 7,10, 0, 9,10, 2, 0,-1,-1,-1,-1},
    { 2, 1,10, 7,11, 6, 8, 0, 3,-1,-1,-1,-1},
    { 1,10, 2, 6, 7,11,-1,-1,-1,-1,-1,-1,-1},
    { 7,11, 6, 3, 9, 1, 3, 8, 9,-1,-1,-1,-1},
    { 9, 1, 0,11, 6, 7,-1,-1,-1,-1,-1,-1,-1},
    { 0, 3, 8,11, 6, 7,-1,-1,-1,-1,-1,-1,-1},
    {11, 6, 7,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    {11, 7, 6,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    { 0, 8, 3,11, 7, 6,-1,-1,-1,-1,-1,-1,-1},
    { 9, 0, 1,11, 7, 6,-1,-1,-1,-1,-1,-1,-1},
    { 7, 6,11, 3, 1, 9, 3, 9, 8,-1,-1,-1,-1},
    { 1, 2,10, 6,11, 7,-1,-1,-1,-1,-1,-1,-1},
    { 2,10, 1, 7, 6,11, 8, 3, 0,-1,-1,-1,-1},
    {11, 7, 6,10, 9, 0,10, 0, 2,-1,-1,-1,-1},
    { 7, 6,11, 3, 2, 8, 8, 2,10, 8,10, 9,-1},
    { 2, 3, 7, 2, 7, 6,-1,-1,-1,-1,-1,-1,-1},
    { 8, 7, 0, 7, 6, 0, 0, 6, 2,-1,-1,-1,-1},
    { 1, 9, 0, 3, 7, 6, 3, 6, 2,-1,-1,-1,-1},
    { 7, 6, 2, 7, 2, 9, 2, 1, 9, 7, 9, 8,-1},
    { 6,10, 7,10, 1, 7, 7, 1, 3,-1,-1,-1,-1},
    { 6,10, 1, 6, 1, 7, 7, 1, 0, 7, 0, 8,-1},
    { 9, 0, 3, 6, 9, 3, 6,10, 9, 6, 3, 7,-1},
    { 6,10, 7, 7,10, 8,10, 9, 8,-1,-1,-1,-1},
    { 8, 4, 6, 8, 6,11,-1,-1,-1,-1,-1,-1,-1},
    {11, 3, 6, 3, 0, 6, 6, 0, 4,-1,-1,-1,-1},
    { 0, 1, 9, 4, 6,11, 4,11, 8,-1,-1,-1,-1},
    { 1, 9, 4,11, 1, 4,11, 3, 1,11, 4, 6,-1},
    {10, 1, 2,11, 8, 4,11, 4, 6,-1,-1,-1,-1},
    {10, 1, 2,11, 3, 6, 6, 3, 0, 6, 0, 4,-1},
    { 0, 2,10, 0,10, 9, 4,11, 8, 4, 6,11,-1},
    { 2,11, 3, 6, 9, 4, 6,10, 9,-1,-1,-1,-1},
    { 3, 8, 2, 8, 4, 2, 2, 4, 6,-1,-1,-1,-1},
    { 2, 0, 4, 2, 4, 6,-1,-1,-1,-1,-1,-1,-1},
    { 1, 9, 0, 3, 8, 2, 2, 8, 4, 2, 4, 6,-1},
    { 9, 4, 1, 1, 4, 2, 4, 6, 2,-1,-1,-1,-1},
    { 8, 4, 6, 8, 6, 1, 6,10, 1, 8, 1, 3,-1},
    { 1, 0,10,10, 0, 6, 0, 4, 6,-1,-1,-1,-1},
    { 8, 0, 3, 9, 6,10, 9, 4, 6,-1,-1,-1,-1},
    {10, 4, 6,10, 9, 4,-1,-1,-1,-1,-1,-1,-1},
    { 9, 5, 4, 7, 6,11,-1,-1,-1,-1,-1,-1,-1},
    { 4, 9, 5, 3, 0, 8,11, 7, 6,-1,-1,-1,-1},
    { 6,11, 7, 4, 0, 1, 4, 1, 5,-1,-1,-1,-1},
    { 6,11, 7, 4, 8, 5, 5, 8, 3, 5, 3, 1,-1},
    { 6,11, 7, 1, 2,10, 9, 5, 4,-1,-1,-1,-1},
    {11, 7, 6, 8, 3, 0, 1, 2,10, 9, 5, 4,-1},
    {11, 7, 6,10, 5, 2, 2, 5, 4, 2, 4, 0,-1},
    { 7, 4, 8, 2,11, 3,10, 5, 6,-1,-1,-1,-1},
    { 4, 9, 5, 6, 2, 3, 6, 3, 7,-1,-1,-1,-1},
    { 9, 5, 4, 8, 7, 0, 0, 7, 6, 0, 6, 2,-1},
    { 4, 0, 1, 4, 1, 5, 6, 3, 7, 6, 2, 3,-1},
    { 7, 4, 8, 5, 2, 1, 5, 6, 2,-1,-1,-1,-1},
    { 4, 9, 5, 6,10, 7, 7,10, 1, 7, 1, 3,-1},
    { 5, 6,10, 0, 9, 1, 8, 7, 4,-1,-1,-1,-1},
    { 5, 6,10, 7, 0, 3, 7, 4, 0,-1,-1,-1,-1},
    {10, 5, 6, 4, 8, 7,-1,-1,-1,-1,-1,-1,-1},
    { 5, 6, 9, 6,11, 9, 9,11, 8,-1,-1,-1,-1},
    { 0, 9, 5, 0, 5, 3, 3, 5, 6, 3, 6,11,-1},
    { 0, 1, 5, 0, 5,11, 5, 6,11, 0,11, 8,-1},
    {11, 3, 6, 6, 3, 5, 3, 1, 5,-1,-1,-1,-1},
    { 1, 2,10, 5, 6, 9, 9, 6,11, 9,11, 8,-1},
    { 1, 0, 9, 6,10, 5,11, 3, 2,-1,-1,-1,-1},
    { 6,10, 5, 2, 8, 0, 2,11, 8,-1,-1,-1,-1},
    { 3, 2,11,10, 5, 6,-1,-1,-1,-1,-1,-1,-1},
    { 9, 5, 6, 3, 9, 6, 3, 8, 9, 3, 6, 2,-1},
    { 5, 6, 9, 9, 6, 0, 6, 2, 0,-1,-1,-1,-1},
    { 0, 3, 8, 2, 5, 6, 2, 1, 5,-1,-1,-1,-1},
    { 1, 6, 2, 1, 5, 6,-1,-1,-1,-1,-1,-1,-1},
    {10, 5, 6, 9, 3, 8, 9, 1, 3,-1,-1,-1,-1},
    { 0, 9, 1, 5, 6,10,-1,-1,-1,-1,-1,-1,-1},
    { 8, 0, 3,10, 5, 6,-1,-1,-1,-1,-1,-1,-1},
    {10, 5, 6,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    {11, 7, 5,11, 5,10,-1,-1,-1,-1,-1,-1,-1},
    { 3, 0, 8, 7, 5,10, 7,10,11,-1,-1,-1,-1},
    { 9, 0, 1,10,11, 7,10, 7, 5,-1,-1,-1,-1},
    { 3, 1, 9, 3, 9, 8, 7,10,11, 7, 5,10,-1},
    { 2,11, 1,11, 7, 1, 1, 7, 5,-1,-1,-1,-1},
    { 0, 8, 3, 2,11, 1, 1,11, 7, 1, 7, 5,-1},
    { 9, 0, 2, 9, 2, 7, 2,11, 7, 9, 7, 5,-1},
    {11, 3, 2, 8, 5, 9, 8, 7, 5,-1,-1,-1,-1},
    {10, 2, 5, 2, 3, 5, 5, 3, 7,-1,-1,-1,-1},
    { 5,10, 2, 8, 5, 2, 8, 7, 5, 8, 2, 0,-1},
    { 9, 0, 1,10, 2, 5, 5, 2, 3, 5, 3, 7,-1},
    { 1,10, 2, 5, 8, 7, 5, 9, 8,-1,-1,-1,-1},
    { 1, 3, 7, 1, 7, 5,-1,-1,-1,-1,-1,-1,-1},
    { 8, 7, 0, 0, 7, 1, 7, 5, 1,-1,-1,-1,-1},
    { 0, 3, 9, 9, 3, 5, 3, 7, 5,-1,-1,-1,-1},
    { 9, 7, 5, 9, 8, 7,-1,-1,-1,-1,-1,-1,-1},
    { 4, 5, 8, 5,10, 8, 8,10,11,-1,-1,-1,-1},
    { 3, 0, 4, 3, 4,10, 4, 5,10, 3,10,11,-1},
    { 0, 1, 9, 4, 5, 8, 8, 5,10, 8,10,11,-1},
    { 5, 9, 4, 1,11, 3, 1,10,11,-1,-1,-1,-1},
    { 8, 4, 5, 2, 8, 5, 2,11, 8, 2, 5, 1,-1},
    { 3, 2,11, 1, 4, 5, 1, 0, 4,-1,-1,-1,-1},
    { 9, 4, 5, 8, 2,11, 8, 0, 2,-1,-1,-1,-1},
    {11, 3, 2, 9, 4, 5,-1,-1,-1,-1,-1,-1,-1},
    { 3, 8, 4, 3, 4, 2, 2, 4, 5, 2, 5,10,-1},
    {10, 2, 5, 5, 2, 4, 2, 0, 4,-1,-1,-1,-1},
    { 0, 3, 8, 5, 9, 4,10, 2, 1,-1,-1,-1,-1},
    { 2, 1,10, 9, 4, 5,-1,-1,-1,-1,-1,-1,-1},
    { 4, 5, 8, 8, 5, 3, 5, 1, 3,-1,-1,-1,-1},
    { 5, 0, 4, 5, 1, 0,-1,-1,-1,-1,-1,-1,-1},
    { 3, 8, 0, 4, 5, 9,-1,-1,-1,-1,-1,-1,-1},
    { 9, 4, 5,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    { 7, 4,11, 4, 9,11,11, 9,10,-1,-1,-1,-1},
    { 3, 0, 8, 7, 4,11,11, 4, 9,11, 9,10,-1},
    {11, 7, 4, 1,11, 4, 1,10,11, 1, 4, 0,-1},
    { 8, 7, 4,11, 1,10,11, 3, 1,-1,-1,-1,-1},
    { 2,11, 7, 2, 7, 1, 1, 7, 4, 1, 4, 9,-1},
    { 3, 2,11, 4, 8, 7, 9, 1, 0,-1,-1,-1,-1},
    { 7, 4,11,11, 4, 2, 4, 0, 2,-1,-1,-1,-1},
    { 2,11, 3, 7, 4, 8,-1,-1,-1,-1,-1,-1,-1},
    { 2, 3, 7, 2, 7, 9, 7, 4, 9, 2, 9,10,-1},
    { 4, 8, 7, 0,10, 2, 0, 9,10,-1,-1,-1,-1},
    { 2, 1,10, 0, 7, 4, 0, 3, 7,-1,-1,-1,-1},
    {10, 2, 1, 8, 7, 4,-1,-1,-1,-1,-1,-1,-1},
    { 9, 1, 4, 4, 1, 7, 1, 3, 7,-1,-1,-1,-1},
    { 1, 0, 9, 8, 7, 4,-1,-1,-1,-1,-1,-1,-1},
    { 3, 4, 0, 3, 7, 4,-1,-1,-1,-1,-1,-1,-1},
    { 8, 7, 4,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    { 8, 9,10, 8,10,11,-1,-1,-1,-1,-1,-1,-1},
    { 0, 9, 3, 3, 9,11, 9,10,11,-1,-1,-1,-1},
    { 1,10, 0, 0,10, 8,10,11, 8,-1,-1,-1,-1},
    {10, 3, 1,10,11, 3,-1,-1,-1,-1,-1,-1,-1},
    { 2,11, 1, 1,11, 9,11, 8, 9,-1,-1,-1,-1},
    {11, 3, 2, 0, 9, 1,-1,-1,-1,-1,-1,-1,-1},
    {11, 0, 2,11, 8, 0,-1,-1,-1,-1,-1,-1,-1},
    {11, 3, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    { 3, 8, 2, 2, 8,10, 8, 9,10,-1,-1,-1,-1},
    { 9, 2, 0, 9,10, 2,-1,-1,-1,-1,-1,-1,-1},
    { 8, 0, 3, 1,10, 2,-1,-1,-1,-1,-1,-1,-1},
    {10, 2, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    { 8, 1, 3, 8, 9, 1,-1,-1,-1,-1,-1,-1,-1},
    { 9, 1, 0,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    { 8, 0, 3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
    {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}};

int polygonise() { 
//																										marching cube algorithm
	int i, j, k, m, n;
	int a, b, c;
	int cell, hash_index, hash_found;
	float step = (parameters.max - parameters.min) / (parameters.step - 1);
	float scale = 1.0 / parameters.precision;
//																										build the grid	
	for(i = 0; i < parameters.step - 1; i++) {		
		for(j = 0; j < parameters.step - 1; j++) {
			for(k = 0; k < parameters.step - 1; k++) {
				cube_index = 0;
//																										one cube at a time				
				cube.p[0].x = parameters.min + (step * i); cube.p[0].y = parameters.min + (step * j); cube.p[0].z = parameters.min + (step * k); 				
				cube.p[1].x = parameters.min + (step * (i+1)); cube.p[1].y = cube.p[0].y ; cube.p[1].z = cube.p[0].z; 
				cube.p[2].x = cube.p[1].x; cube.p[2].y = parameters.min + (step * (j+1)); cube.p[2].z = cube.p[0].z; 
				cube.p[3].x = cube.p[0].x; cube.p[3].y = cube.p[2].y; cube.p[3].z = cube.p[0].z;
				cube.p[4].x = cube.p[0].x; cube.p[4].y = cube.p[0].y; cube.p[4].z = parameters.min + (step * (k+1));
				cube.p[5].x = cube.p[1].x; cube.p[5].y = cube.p[0].y; cube.p[5].z = cube.p[4].z;
				cube.p[6].x = cube.p[1].x; cube.p[6].y = cube.p[2].y; cube.p[6].z = cube.p[4].z;
				cube.p[7].x = cube.p[0].x; cube.p[7].y = cube.p[2].y; cube.p[7].z = cube.p[4].z;
//																										compute the function value at the cube vertices
				for(cube_vertex = 0; cube_vertex < 8; cube_vertex++) { 
					stack_index = -1;
					for(token = 0; token < equation.xyz_tokennumber; token++) {
						switch(equation.xyz_token[token].type) {
							case NUMBER_TYPE:	execution_stack[++stack_index] = equation.xyz_token[token].value;
												break;
								
							case UNKNOWN_TYPE:	if(equation.xyz_token[token].index == X) execution_stack[++stack_index] = cube.p[cube_vertex].x;
												else if(equation.xyz_token[token].index == Y) execution_stack[++stack_index] = cube.p[cube_vertex].y;
												else execution_stack[++stack_index] = cube.p[cube_vertex].z;
												break;
								
							case VARIABLE_TYPE:	execution_stack[++stack_index] = variable.value[equation.xyz_token[token].index];	
												break;
								
							case OPERATOR_TYPE:	execution_stack[stack_index] = sqrt(execution_stack[stack_index]);
												break;
												
							case FUNCTION_TYPE:if(function[equation.xyz_token[token].index].argnum == 1)
													execution_stack[stack_index] = 
														function[equation.xyz_token[token].index].function(execution_stack[stack_index]);
												else {	
													execution_stack[stack_index - 1] = 
													function[equation.xyz_token[token].index].function(execution_stack[stack_index - 1],execution_stack[stack_index]);
													--stack_index;
												}
												break;		
						}
					}
//																											compare to the isovalue
					if((cube.val[cube_vertex] = execution_stack[stack_index]) < parameters.isovalue) cube_index |= (1 << cube_vertex);			
				}
//																											interpolate the intersection points
				edge_index = edge_table[cube_index];
				for(interpolate = 0; interpolate < 12; interpolate++) {
					if(!(edge_index & (1 << interpolate))) continue;
					int i0 = interpolate_vertex[interpolate][0]; int i1 = interpolate_vertex[interpolate][1];					
					if (fabs(parameters.isovalue - cube.val[i0]) < interpolate_delta)	{
						cube_edge[interpolate].x = cube.p[i0].x; cube_edge[interpolate].y = cube.p[i0].y; cube_edge[interpolate].z = cube.p[i0].z;}
					else if (fabs(parameters.isovalue - cube.val[i1]) < interpolate_delta) {
						cube_edge[interpolate].x = cube.p[i1].x; cube_edge[interpolate].y = cube.p[i1].y; cube_edge[interpolate].z = cube.p[i1].z;}
					else if (fabs(cube.val[i0] - cube.val[i1]) < interpolate_delta) {
						cube_edge[interpolate].x = cube.p[i0].x; cube_edge[interpolate].y = cube.p[i0].y; cube_edge[interpolate].z = cube.p[i0].z;}
					else {
						float mu = (parameters.isovalue - cube.val[i0]) / (cube.val[i1] - cube.val[i0]);
						cube_edge[interpolate].x = cube.p[i0].x + mu * (cube.p[i1].x - cube.p[i0].x);
						cube_edge[interpolate].y = cube.p[i0].y + mu * (cube.p[i1].y - cube.p[i0].y);
						cube_edge[interpolate].z = cube.p[i0].z + mu * (cube.p[i1].z - cube.p[i0].z);
					}
				}
//																											generate the triangles
				for (m = 0; triangle_table[cube_index][m] != -1; m += 3) {
					for(n = 0; n < 3; n++) {
						if(num_poly_verts >= MAX_HASH_INDEX){hpg_clear(); hpg_draw_text("too many triangles!", 0, 0); hpg_flip(); keyb_getkey(1); return 1;}		
						idx = triangle_table[cube_index][m+n];
//																											spatial partitioning for vertex clustering
						a = floor(cube_edge[idx].x * scale); b = floor(cube_edge[idx].y * scale); c = floor(cube_edge[idx].z * scale);
						if( (hash_index = (a * 17 + b * 101 + c) % HASH_TABLE_SIZE) < 0) hash_index += HASH_TABLE_SIZE;
						hash_found = 0;HASH_VERTEX *p; 
						for (cell = 0; cell < hash_table[hash_index].num_cell; cell++) {
							p = &hash_table[hash_index].vertex[cell];
							if (p->a == a && p->b == b && p->c == c) {
								p->x += cube_edge[idx].x; p->y += cube_edge[idx].y;
								p->z += cube_edge[idx].z; p->num_hash_vertex++; 
								polys_vertex_index[num_poly_verts++] = p; hash_found = 1; break;
							}
						}	
						if(!hash_found) {
							if(hash_table[hash_index].num_cell < MAX_HASH_VERTEX) {
								p = &hash_table[hash_index].vertex[hash_table[hash_index].num_cell++];
								p->x = cube_edge[idx].x; p->y = cube_edge[idx].y; p->z = cube_edge[idx].z;
								p->a = a; p->b = b; p->c = c; p->num_hash_vertex++; 
								polys_vertex_index[num_poly_verts++] = p; num_verts++;
							}
							else {hpg_clear(); hpg_draw_text("collision overflow", 0, 0); hpg_flip(); keyb_getkey(1); return 1;}
						}
					}	
				}		
			}	
		}
	}
	return 1;
}
	
int get_equation() {
	if(	(equation.expression = sat_stack_pop_string_alloc()) == NULL ) { printf("Error: equation missing\n"); keyb_getkey(1); return 0;}
	equation_info.xyz = sat_stack_pop_string_alloc(); equation_info.xyz[strlen(equation_info.xyz) - 1] = '\0';
	if(parse_expression() == NULL) return 0; get_parameters(); get_variable();	
	if(parameters.isovalue <= -2.0) parameters.isovalue = -2.0; else if(parameters.isovalue >= 2.0) parameters.isovalue = 2.0;
	if(parameters.step <= 7.0) parameters.step = 7.0; else if(parameters.step >= 15.0) parameters.step = 15.0;
	if(parameters.min <= -5.0) parameters.min = -5.0; else if(parameters.min >= 10.0) parameters.min = 10.0; 
	if(parameters.max <= -5.0) parameters.max = -5.0; else if(parameters.max >= 10.0) parameters.max = 10.0; 
	if(parameters.scale <= 0.0) parameters.scale = 0.5; else if(parameters.scale >= 10.0) parameters.scale = 10.0; 
	if(parameters.precision <= 0.2) parameters.precision = 0.2; else if(parameters.precision >= 2.) parameters.precision = 2.; 
	if(parameters.max <= parameters.min) parameters.max = parameters.min + 2.0; return 1;
}	

int parse_expression() {
	char *expression = equation.expression;

	if(*expression != '{' || expression[strlen(expression) - 1] != '}') 
		{ printf("Error: not a valid expression\n"); keyb_getkey(1); return 0;}
	
	expression += 2; equation.xyz_tokennumber = 0;
	
	while(*expression != '}') {
	
		if(equation.xyz_tokennumber >= MAX_TOKEN)  
			{printf("Error: too many tokens!\n"); keyb_getkey(1); return 0;}

		if( isnumber(*expression) || ( *expression == '-' &&  isnumber(* (expression + 1)))) {
			equation.xyz_token[equation.xyz_tokennumber].value = atof_local(expression);
			equation.xyz_token[equation.xyz_tokennumber++].type = NUMBER_TYPE;
			while(*expression != ' ') ++expression; ++expression;
		}
		
		else if( isvariable(*expression) && *(expression + 1) == ' ') {
			int index = *expression - 'a';
			equation.xyz_token[equation.xyz_tokennumber].index = index;
			equation.xyz_token[equation.xyz_tokennumber++].type = VARIABLE_TYPE;
			variable.flag[index] = 1; expression += 2;
		}
	
		else if( isunknown(*expression) && *(expression + 1) == ' ') {
			equation.xyz_token[equation.xyz_tokennumber].index = *expression - 'x';
			equation.xyz_token[equation.xyz_tokennumber++].type = UNKNOWN_TYPE;
			expression += 2;
		}

		else if( isoperator((int) *expression) && *(expression + 1) == ' ') {
			equation.xyz_token[equation.xyz_tokennumber].index = (int) *expression;
			equation.xyz_token[equation.xyz_tokennumber++].type = OPERATOR_TYPE;
			expression += 2;
		}

		else {
			int i = 0;
			while(*expression != ' ') function_name[i++] = *expression++;
			function_name[i] = '\0'; ++expression;	
			int found = 0;
			for(i = 0; i < MAX_FUNCTION; i++) {
				if(!strcmp(function[i].name, function_name)) {
					equation.xyz_token[equation.xyz_tokennumber].index = i;
					equation.xyz_token[equation.xyz_tokennumber++].type = FUNCTION_TYPE;
					found = 1;break;
				}
			}			
			if(!found) {printf("Error: token unknown in expression: %s\n", function_name); keyb_getkey(1); return 0;}
		}	
	}
	return 1;
}


void get_parameters(){
	int i = 0;
	if( (equation.parameter = sat_stack_pop_string_alloc()) ) {
		if(*equation.parameter == '{' && equation.parameter[strlen(equation.parameter) - 1] == '}') {
			equation.parameter += 2;
			while(*equation.parameter != '}' && i < MAX_PARAMETER) {
				if( isnumber(*equation.parameter) || ( *equation.parameter == '-' && isnumber(* (equation.parameter + 1)))) {
					parameters.p[i++] = atof_local(equation.parameter); 
					while(*equation.parameter != ' ') ++equation.parameter; ++equation.parameter;
				}
				else return;
			}
		}
	}
}
void get_variable(){
	int i = 0;
	if( (equation.variable = sat_stack_pop_string_alloc()) ) {
		if(*equation.variable == '{' && equation.variable[strlen(equation.variable) - 1] == '}') {
			equation.variable += 2;		
			while(*equation.variable != '}' && i < MAX_VARIABLE) {
				if( isnumber(*equation.variable) || ( *equation.variable == '-' && isnumber(* (equation.variable + 1)))) {
					variable.value[i++] = atof_local(equation.variable);
					while(*equation.variable != ' ') ++equation.variable; ++equation.variable;
				}
				else return;
			}
		}
	}
}
//
double addition() {return execution_stack[stack_index - 1] + execution_stack[stack_index];};
double subtraction() {return execution_stack[stack_index - 1] - execution_stack[stack_index];};
double multiplication() {return execution_stack[stack_index - 1] * execution_stack[stack_index];};
double division() {
	if(execution_stack[stack_index] == 0.0) execution_stack[stack_index] = EPSILON_E5;
	return execution_stack[stack_index - 1] / execution_stack[stack_index];
}
double neg(){return -execution_stack[stack_index];}
double inv(){
	if(execution_stack[stack_index] == 0.0) execution_stack[stack_index] = EPSILON_E5;
	return 1.0 / execution_stack[stack_index];
}
double sq() { return execution_stack[stack_index] * execution_stack[stack_index];};
double sinhp(){return (exp(execution_stack[stack_index]) - exp(-execution_stack[stack_index])) * 0.5;}
double coshp(){return (exp(execution_stack[stack_index]) + exp(-execution_stack[stack_index])) * 0.5;}
double tanhp(){double q; if( (q = coshp()) == 0.0) q = EPSILON_E5; return (sinhp() / q);}

double atof_local(const char * _buf) {
    double val,expo;
    int minus,fraction_digits;
    // changed due to certain overflow conditions (excessive long strings)
    // <ibl ; 2008-04-14>
    /*ULONGLONG */ double integral,ifraction;

    char c,*__buf,*buf;
	
	if( (__buf = (char *)  malloc((strlen(_buf) + 1) * sizeof(char)) ) == NULL ) 
		return(error_memory("atof_local\n"));
    
    buf = lower(__buf,_buf); // just handle lowercase chars

    integral = ifraction = 0;
    fraction_digits = 0;
    val = 0.0;
    expo = 1.0;
    
    if((minus = (*buf == '-')))
        buf++;
    if (*buf == '+')
        buf++;
        
    while (*buf && *buf != '.') {
        c = *buf++;
        if (isdigit(c))
            integral = 10 * integral + c - '0';
        else {
            
            if (c == 'e') {
                buf--;
                goto parse_expo;
            }
            goto finish;
        }
    }
    
    
    if (*buf == '.') {
        buf++;
        while (*buf && *buf != 'e' ) {
            c = *buf++;
            if (! isdigit(c)) goto finish;
            fraction_digits++;
            ifraction = 10 * ifraction + c - '0';
        }
    }
    
parse_expo:
    
    if (*buf == 'e') {
        int expo_minus,ee = 0;
        buf++;
        if ((expo_minus = (*buf == '-')))
            buf++;
        if (*buf == '+')
            buf++;
        while (*buf) {
            c = *buf++;
            if (! isdigit(c))
                goto finish;
            ee = 10 * ee + c - '0';
        }
		// REPLACED IPOW(10,X) BY I10POWX(X) - CLAUDIO MAR 2006
		
        expo = i10powx(expo_minus ? -ee : ee);
    }

finish:
	// REPLACED IPOW(10,X) BY I10POWX(X) - CLAUDIO MAR 2006
	
	free(__buf);
	
    val = expo * ((double) integral + (double) ifraction / i10powx(fraction_digits));
    return minus ? -val : val;
}






//																								routines
int compare_polygone(const void *arg1, const void *arg2) {
	float z1, z2; POLYGONE_PTR poly_1, poly_2;
	poly_1 = *((POLYGONE_PTR *)(arg1)); poly_2 = *((POLYGONE_PTR *)(arg2));
	z1 = (float)0.33333*(surface.vlist[poly_1->vert[0]].z + surface.vlist[poly_1->vert[1]].z + surface.vlist[poly_1->vert[2]].z);
	z2 = (float)0.33333*(surface.vlist[poly_2->vert[0]].z + surface.vlist[poly_2->vert[1]].z + surface.vlist[poly_2->vert[2]].z);
	if (z1 > z2) return(-1); else if (z1 < z2) return(1); else return(0);
} 

void init_plane(PLANE_PTR plane, POINT3D_PTR p0, VECTOR3D_PTR normal) {
	plane->p0.x = p0->x;plane->p0.y = p0->y;plane->p0.z = p0->z;								
	float length = (float) sqrt( normal->x * normal->x + normal->y * normal->y + normal->z * normal->z);
	if (length < EPSILON_E5) return;							
	plane->n.x = normal->x / length; plane->n.y = normal->y / length; plane->n.z = normal->z / length;
} 

void init_matrice(MATRICE_PTR ma,	float m00, float m01, float m02, float m03, float m10, float m11, float m12, float m13,
										float m20, float m21, float m22, float m23, float m30, float m31, float m32, float m33) {
	ma->M00 = m00; ma->M01 = m01; ma->M02 = m02; ma->M03 = m03;
	ma->M10 = m10; ma->M11 = m11; ma->M12 = m12; ma->M13 = m13;
	ma->M20 = m20; ma->M21 = m21; ma->M22 = m22; ma->M23 = m23;
	ma->M30 = m30; ma->M31 = m31; ma->M32 = m32; ma->M33 = m33;
} 

void multiply_matrice(MATRICE_PTR ma, MATRICE_PTR mb, MATRICE_PTR mprod) {
	int row,col,index;
	for (row=0; row<4; row++) {
		for (col=0; col<4; col++) {
			float sum = 0; 
			for (index = 0; index < 4; index++) {
				sum += (ma->M[row][index] * mb->M[index][col]);
			} 
			mprod->M[row][col] = sum;
        } 
    } 
} 

void build_rotation_matrice(float theta_x, float theta_y, float theta_z, MATRICE_PTR mrot)  {
	MATRICE mx, my, mz, mtmp;       							
	float sin_theta = 0, cos_theta = 0;						   	  							
	MAT_IDENTITY(mrot);										
	cos_theta = cos(theta_x);
	sin_theta = sin(theta_x);
	init_matrice(&mx, 	1,    0,         0,         0,
						0,    cos_theta, sin_theta, 0,
						0,   -sin_theta, cos_theta, 0,
						0,    0,         0,         1);
	cos_theta = cos(theta_y);
	sin_theta = sin(theta_y);
	init_matrice(&my,	cos_theta, 0, -sin_theta, 0,  
						0,         1,  0,         0,
						sin_theta, 0,  cos_theta,  0,
						0,         0,  0,          1);
	cos_theta = cos(theta_z);
	sin_theta = sin(theta_z);
	init_matrice(&mz, 	cos_theta, 	sin_theta, 0, 0,  
						-sin_theta, cos_theta, 0, 0,
						0,         	0,         1, 0,
						0,         	0,         0, 1);
	multiply_matrice(&mx, &my, &mtmp);
	multiply_matrice(&mtmp, &mz, mrot);
}

void init_camera(CAMERA_PTR cam, POINT4D_PTR cam_pos, VECTOR4D_PTR cam_dir, float near_z, float far_z, float fov, float viewport_width, float viewport_height) {
		cam->pos.x = cam_pos->x;cam->pos.y = cam_pos->y;cam->pos.z = cam_pos->z;cam->pos.w = cam_pos->w;
		cam->dir.x = cam_dir->x;cam->dir.y = cam_dir->y;cam->dir.z = cam_dir->z;cam->dir.w = cam_dir->w;
		cam->near_z = near_z; cam->far_z  = far_z;     	
		cam->viewport_width  = viewport_width; cam->viewport_height = viewport_height;
		cam->viewport_center_x = (viewport_width-1)/2; cam->viewport_center_y = (viewport_height-1)/2;
		cam->aspect_ratio = (float) viewport_width / (float) viewport_height;
		MAT_IDENTITY(&cam->mcam); MAT_IDENTITY(&cam->mper); MAT_IDENTITY(&cam->mscr);
		cam->fov = fov; cam->viewplane_width  = 2.0; cam->viewplane_height = 2.0/cam->aspect_ratio;
		float tan_fov_div2 = tan(DEG_TO_RAD(fov / 2));
		cam->view_dist = (0.5) * (cam->viewplane_width) * tan_fov_div2;
		POINT3D pt_origin; pt_origin.x = 0.0; pt_origin.y = 0.0; pt_origin.z = 0.0;
		VECTOR3D vn; 
		vn.x =  1.0;vn.y =  0.0;vn.z = -1.0; init_plane(&cam->rt_plane, &pt_origin, &vn);
		vn.x = -1.0;vn.y =  0.0;vn.z = -1.0; init_plane(&cam->lt_plane, &pt_origin, &vn);
		vn.x =  0.0;vn.y =  1.0;vn.z = -1.0; init_plane(&cam->tp_plane, &pt_origin, &vn);
		vn.x =  0.0;vn.y = -1.0;vn.z = -1.0; init_plane(&cam->bt_plane, &pt_origin, &vn); 		
}









